/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#ifndef PKT_HPP_
#define PKT_HPP_

#include <vector>
#include "common.hpp"

#ifdef USE_ORION
extern "C" {
#include "SIM_power.h"
#include "SIM_router_power.h"
#include "SIM_power_router.h"
}
#endif 

using namespace std;

typedef class Flit {
private:
    class Packet * pkt;       // which packet does it belong to

    unsigned int seq_id;               // it's the seq_id-th flit of the packet
    unsigned int flit_size;            // size of the flit
    unsigned int pkt_size;             // how many flits in the same packet?

    // only header flit will contain src_id and dst_id.
    int src_id;
    int dst_id;

    int control_flag;         // reserved

    int * payload;           
    Atom_type raw_data;

    bool processed;           // whether it has been processed by the current node
    int arrival_timestamp;    // when did it arrive at the current node

    vector<class Router *> routing_path;    // records the routers on the path it passes

    Atom_type convert_to_raw(void);

public:
    Flit(class Packet * p, unsigned int s_id, unsigned int pkt_sz, 
         unsigned int flit_sz, int * pld = 0);
    ~Flit();

    int get_src_id(void) const;
    Position get_src_position(void) const;
    void write_src_id(int a_id) { src_id = a_id; }
    int get_dst_id(void) const;
    Position get_dst_position(void) const;
    void write_dst_id(int a_id) { dst_id = a_id; }

    unsigned int get_packet_size(void) const;
    unsigned int get_flit_size(void) const { return flit_size; }

    int * set_payload(int * pl);
    int * get_payload(void) { return payload; }

    class Packet * get_packet(void) { return pkt; }

    bool is_header(void) const { return (seq_id == 0); }
    bool is_tail(void) const { return (seq_id + 1 == pkt_size); }
    bool arrived_in_this_cycle(void) const;
    void arrive(void);
    bool is_processed(void) const { return processed; }
    void reset_processed_flag(void) { processed = false; }
    void get_processed(void) { processed = true; }

    void add_a_hop(class Router * router); 
    class Router * get_previous_hop(void);

    Atom_type raw(void) const { return raw_data; }     // retrieve the raw data of the flit to feed to power module

} Flit, *pFlit;


typedef class Packet {
    //    static int id_base;
    int id;                       // an id for this packet, (source, id) identifies a unique packet)
    vector<class Flit *> flits;
    unsigned int packet_size;     // how many flits in this packet
    unsigned int flit_size;

    int birth_time;               // when this packet is borned
    int death_time;               // when this packet die

public:
    Packet(int pkt_id, unsigned int pkt_sz, unsigned int flit_sz);
    //    Packet(vector<class Flit *> fs);
    ~Packet();

    pFlit get_flit(int f_id);
    bool add_flit(class Flit * a_flit);
    unsigned int size(void) const { return packet_size; }
    int get_id(void) const { return id; }
    void set_death_time(int timestamp) { death_time = timestamp; }
    int get_latency(void) const { return death_time - birth_time; }
    Position get_src_position(void) const { return flits[0]->get_src_position(); }
    Position get_dst_position(void) const { return flits[0]->get_dst_position(); }
} Packet, *pPacket;


typedef vector<pFlit>::iterator iter_flit;
typedef vector<pPacket>::iterator iter_packet;


#endif // PKT_HPP_
