/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#include "common.hpp"
#include "router.hpp"
#include "msg.hpp"
#include "global_val.hpp"
#include <iostream>

Router::Router(pNetwork n, Position p) : Addressee(p), 
                                         sw(param.n_of_switch_ports) ,
                                         arbiter(param.arbitration_delay, 
                                                 param.n_of_switch_ports) {
    if (param.extreme_verbose) 
        cout << "[I] Building router for location " << pos << endl;

    network = n;

    in_ports.clear();
    out_ports.clear();

    for (unsigned int i=0; i<network->get_num_of_ports(); i++) {
        Direction d = (Direction) i;
        pIn_port a_in_port = new In_port(this, d); 
        pOut_port a_out_port = new Out_port(this, d);
        in_ports.push_back(a_in_port);
        out_ports.push_back(a_out_port);
    }
    arbiter.bind(this);
    sw.bind(this);

    // bind switch ends with the input/output channels
    assert(sw.get_size() == in_ports.size() * in_ports[0]->get_size());
    assert(sw.get_size() == out_ports.size() * out_ports[0]->get_size());

    int index = 0;
    for (vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++) {
        for (unsigned int i=0; i<(*iter)->get_size(); i++) {
            pSw_input_end an_end = sw.get_input_end(index++);
            pInput_channel a_chan = (pInput_channel) (*iter)->get_channel(i);
            an_end->bind(a_chan);
            a_chan->bind(an_end);
        }
    }

    index = 0;
    for (vector<pOut_port>::iterator iter=out_ports.begin(); iter<out_ports.end(); iter++) {
        for (unsigned int i=0; i<(*iter)->get_size(); i++) {
            pSw_output_end an_end = sw.get_output_end(index++);
            pOutput_channel a_chan = (pOutput_channel) (*iter)->get_channel(i);
            an_end->bind(a_chan);
            a_chan->bind(an_end);
        }
    }

    // set up the switch's ebit metric using default value
    sw.set_ebit(param.xbar_ebit);

    // create orion power module
#ifdef USE_ORION
    if (param.orion_power.use_orion_power_model) 
        power_module = new Power_module(param.n_of_ports);
    else 
        power_module = NULL;
#endif

}


Router::~Router() {
    if (param.extreme_verbose) 
        cout << "[I] Deleting router for location " << pos << endl;

    for (vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++)
        delete (*iter);
    for (vector<pOut_port>::iterator iter=out_ports.begin(); iter<out_ports.end(); iter++)
        delete (*iter);

#ifdef USE_ORION
    if (power_module)
        delete power_module;
#endif
}


void Router::set_input_channel_buffer_size(int s) {
    for (vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++)
        (*iter)->set_channel_buffer_size(s);
}


void Router::set_input_channel_buffer_size(int s, Direction dir) {
    if (dir == local) {
        cerr << "[W] Overriding local input buffer depth is not allowed. Action skipped." << endl;
        return;
    }
    pIn_port p = get_in_port(dir); 
    p->set_channel_buffer_size(s);
}


void Router::set_output_channel_buffer_size(int s) {
    for (vector<pOut_port>::iterator iter=out_ports.begin(); iter<out_ports.end(); iter++)
        (*iter)->set_channel_buffer_size(s);
}


void Router::set_routing_scheme(Routing_scheme rs) {
    for (vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++)
        (*iter)->set_routing_scheme(rs);
}

void Router::set_routing_table(char * table_string) { //UYO
    for(vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++)
        //cout <<"Why not? \n";  
        (*iter)->set_routing_table(table_string);
  
    /* cout <<"@NODE   " << (this->pos.x)*param.n_of_rows+(this->pos.y) <<"\n";
       cout <<"routing_scheme  table ";
       for (int i=0; i<strlen(table_string); i++)
       cout << table_string[i]; //<<" ";
       cout <<"\n";  */
}

// find the index-th input virtual channel in the dir direction of the router
pInput_channel Router::get_in_channel(Direction dir, int index) {
    pIn_port a_port = in_ports[dir];
    return ((pInput_channel) a_port->get_channel(index));
}


pOutput_channel Router::get_out_channel(Direction dir, int index) {
    pOut_port a_port = out_ports[dir];
    return ((pOutput_channel) a_port->get_channel(index));
}


// return the router which is on the dir-side of the router
// return 0 if there does not exist a router (for instance,
// at the edge of the mesh 
pRouter Router::get_router(Direction dir) {
    Position a_pos = pos;
    switch (dir) {
    case north:
        a_pos.y ++;
        break;
    case south:
        a_pos.y --;
        break;
    case east:
        a_pos.x ++;
        break;
    case west:
        a_pos.x --;
        break;
    case local:
        return this;
    default:
        return 0;
    }

    if (network->get_topology() == mesh) {
        if (a_pos.x >= network->get_num_of_cols() || a_pos.x < 0)
            return 0;
        if (a_pos.y >= network->get_num_of_rows() || a_pos.y < 0)
            return 0;
        return network->get_router(a_pos);
    }
    else if (network->get_topology() == torus) {
        // increase to avoid negative numbers
        a_pos.x += network->get_num_of_cols();
        a_pos.y += network->get_num_of_rows();
        
        a_pos.x = a_pos.x % network->get_num_of_cols();
        a_pos.y = a_pos.y % network->get_num_of_rows();
        return network->get_router(a_pos);
    }
    else {
        cerr << ERRO_ARCH_NOT_SUPPORTED;
        return 0;
    }
}


void Router::in_ports_tick(void) {
    for (vector<pIn_port>::iterator iter=in_ports.begin(); iter<in_ports.end(); iter++) 
        (*iter)->tick();
}

void Router::out_ports_tick(void) {
    for (vector<pOut_port>::iterator iter=out_ports.begin(); iter<out_ports.end(); iter++) 
        (*iter)->tick();
}

void Router::arbiter_tick(void) {
    arbiter.tick();
}

